##### Imports #####

import os
from re import S
import math
import streamlit as st
import matplotlib as plt
import pandas as pd
import textwrap
import statsmodels.formula.api as smf
import statsmodels.tsa.api as smt
import statsmodels.api as sm
import scipy.stats as scs
import numba
import numpy as np
import seaborn as sns

import matplotlib.cm as cm
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import colors
import matplotlib.pyplot as plt
import altair as alt

##### Plotly ######
import plotly.express as px
import plotly.graph_objects as go
import plotly.figure_factory as ff
import plotly.graph_objects as go
import chart_studio
from plotly import tools
from plotly.subplots import make_subplots

import time #from datetime import datetime
import datetime


from scipy.fftpack import rfft
from scipy.stats import boxcox
from sklearn.cluster import AgglomerativeClustering, KMeans
from sklearn.metrics import davies_bouldin_score
from sklearn_extra.cluster import KMedoids
from sklearn.metrics import calinski_harabasz_score
from scipy.cluster.hierarchy import single, complete, average, ward, dendrogram, linkage
from sklearn.metrics.cluster import contingency_matrix
from sklearn.manifold import TSNE
from sklearn.preprocessing import MinMaxScaler
from sklearn.cluster import KMeans
from sklearn.impute import SimpleImputer

from datetime import timedelta #from datetime import datetime


# Algorithms
from tslearn.barycenters import dtw_barycenter_averaging
from tslearn.clustering import TimeSeriesKMeans
from sktime.distances import dtw_distance
from dtaidistance import clustering, dtw
#from fcmeans import FCM


# IMplementation for pyclustering kmeans
from pyclustering.cluster.kmeans import kmeans
from pyclustering.cluster.center_initializer import random_center_initializer
from pyclustering.cluster.encoder import type_encoding
from pyclustering.cluster.encoder import cluster_encoder
from pyclustering.utils.metric import distance_metric
from pyclustering.cluster.center_initializer import kmeans_plusplus_initializer
from pyclustering.cluster.fcm import fcm
from sklearn.metrics import pairwise_distances
from validclust import cop, dunn


# Preprocessing
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics.cluster import contingency_matrix
from tslearn.clustering import TimeSeriesKMeans
from tslearn.preprocessing import TimeSeriesScalerMeanVariance
from netdata_pandas.data import get_data, get_chart_list
from am4894plots.plots import plot_lines, plot_lines_grid
from matplotlib.patches import Ellipse
from sklearn import preprocessing

#from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score, silhouette_samples
from yellowbrick.cluster import SilhouetteVisualizer, KElbowVisualizer
from sklearn.model_selection import train_test_split
#from statsmodels.tsa.arima_model import ARIMA

#import warnings  # `do not disturbe` mode

#warnings.filterwarnings('ignore')


import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import silhouette_score, davies_bouldin_score, calinski_harabasz_score
#from tslearn.clustering import TimeSeriesKMeans, KMedoids
from tslearn.metrics import dtw, soft_dtw
from tslearn.utils import to_time_series_dataset
from sklearn.cluster import AgglomerativeClustering
import os

# Helper function to compute SBD (not provided directly by libraries, needs custom implementation)
def sbd_distance(x, y):
    # Placeholder for SBD calculation. Replace with actual SBD computation logic.
    return np.linalg.norm(x - y)

# Define clustering methods and parameters
clustering_methods = {
    "PAM": KMedoids,
    "K-shape": TimeSeriesKMeans,
    "DBA": TimeSeriesKMeans,
    "Single": AgglomerativeClustering,
    "Average": AgglomerativeClustering,
    "Complete": AgglomerativeClustering,
    "Ward": AgglomerativeClustering,
    "Centroid": AgglomerativeClustering,
    "Median": AgglomerativeClustering,
}

distance_metrics = {
    "Euclidean": "euclidean",
    "DTW": dtw,
    "SBD": sbd_distance,
}

# Helper function for evaluating clustering
def evaluate_clustering(labels, data):
    sil = silhouette_score(data, labels, metric="euclidean")
    db = davies_bouldin_score(data, labels)
    ch = calinski_harabasz_score(data, labels)
    return sil, db, ch


# Helper function to preprocess data
def preprocess_data(data):
    """Preprocess data to handle NaN values and ensure valid input for clustering."""
    # Impute NaN values with column mean
    imputer = SimpleImputer(strategy="mean")
    data = imputer.fit_transform(data)
    
    # Check if the data is still valid (non-empty and sufficient samples)
    if data.size == 0 or data.shape[0] < 2:
        raise ValueError("Data is empty or insufficient for clustering.")
    
    return data

# Filter features based on variance
def filter_low_variance_features(data, threshold=0.1):
    """ Filter out features with very low variance to ensure sufficient cluster differentiation. """
    variances = data.var(axis=0)
    selected_features = data.loc[:, variances > threshold]  # Variance above the threshold
    return selected_features

# Process a single file
def process_file(file_path):
    print(f"Processing file: {file_path}")
    data = pd.read_csv(file_path, parse_dates=["Time"])

    # Extract and scale features
    features = data.iloc[:, 1:]  # Select all columns except the 'Time' column

    # Filter out low variance features
    features = filter_low_variance_features(features, threshold=0.1)

    data = preprocess_data(features)  # Handle NaN and preprocessing
    scaler = StandardScaler()
    data_scaled = scaler.fit_transform(data)
    data_ts = to_time_series_dataset(data_scaled)

    results = []

    for method_name, method_class in clustering_methods.items():
        for metric_name, metric in distance_metrics.items():
            try:
                if method_name in ["PAM", "DBA", "K-shape"]:
                    # Partitioning clustering
                    model = method_class(n_clusters=4, metric=metric if method_name != "K-shape" else "softdtw")
                else:
                    # Hierarchical clustering
                    linkage = method_name.lower()
                    model = method_class(n_clusters=4, linkage=linkage)

                # Fit and predict
                if method_name in ["PAM", "DBA", "K-shape"]:
                    labels = model.fit_predict(data_ts.reshape(len(data_ts), -1))
                else:
                    labels = model.fit_predict(data_scaled)

                # Evaluate clustering
                sil, db, ch = evaluate_clustering(labels, data_scaled)
                results.append((method_name, metric_name, sil, db, ch))

                print(f"{method_name} with {metric_name}: Silhouette={sil:.3f}, DB={db:.3f}, CH={ch:.3f}")
            except Exception as e:
                print(f"Error with {method_name} and {metric_name}: {e}")

    return results



# Process all files in a folder
def process_folder(folder_path):
    all_results = []
    for file_name in os.listdir(folder_path):
        if file_name.endswith(".csv"):
            file_path = os.path.join(folder_path, file_name)
            results = process_file(file_path)
            all_results.extend([(file_name, *result) for result in results])

    # Save results to a CSV file
    results_df = pd.DataFrame(all_results, columns=["File", "Method", "Metric", "Silhouette", "Davies-Bouldin", "Calinski-Harabasz"])
    results_df.to_csv("clustering_results.csv", index=False)
    print("Results saved to clustering_results.csv")


# Main entry point
if __name__ == "__main__":
    folder_path = "/home/jvaldes/Desktop/krishna-thesis/thesiscodeimplementation/data"  # Replace with your folder path
    process_folder(folder_path)

